function mapping(action)
%% This M-file finds the bounds for a magnitude/phase stability diagram.
%
% OUTPUTS: - NTR  : Numer of unstable area's (1x2)
%          - Mlow : Lower bound of the magnitude (1x2)
%          - Plow : Lower bound of the phase (1x2)
%          - Mupp : Upper bound of the magnitude (1x2)
%          - Pupp : Upper bound of the phase (1x2)
%          - Nst  : A measure for the amount of conservatism added.
%                   If Nst = 100%, then ARinn = ARout: there is no
%                   uncertainty and the outer bounds equal the inner
%                   bounds!
%          - rev  : Indicates whether the algorithm has 'reversed'
%                   direction.
%                   rev = -2: there are no stable/passive combinations,
%                   rev = -1: there are only stable/passive combinations.
%                   rev = 1 : there are stable/passive and unstable/active combinations.
%                             However, the unstable area cannot be bounded (there is no
%                             magnitude for which all phase points are
%                             stable/passive and no phase for which all
%                             magnitude points are stable/passive).
%                   rev = 0 : indicates there are stable/passive and
%                               unstable/active combinations that can be bounded
%                               (=normal procedure).
%
% INPUTS:  - crit  : Stabilitygrid (1:Stable 0:Unstable) (MagxPhase)
%          - Mag    : Magnitude vector (1xnMag)
%          - Phase    : Phase vector (1xnPhase)
%
% Three approaches are possible
% Approach apprch = 0
%       There are stable/passive combinations, the aim is to find
%       unconditionnaly stable/passive magnitude and phase.
% Approach apprch = 1
%       Approach 'apprch = 0' failed! There are stable/passive combinations
%       but no unconditionally stable/passive magnitude or phase exists.
%       The aim is to find unconditionnaly unstable/active magnitude and
%       phase. Therefore, the algorithm is reversed: instead of bounding
%       the unstable/active area, the algorithm is bounding the stable/passive
%       area. If the algorithm is reversed, a black line in the bottom of
%       the bode diagram will be shown.
% Approach apprch = 2
%       There are no or only stable/passive combinations so skip everything.
%
% The magnitude/phase grid can only contain 1 left and 1 right bound
% on the magnitude. On the phase it can contain more bounds. For now.
%
%%
global data

Freq = data.parameters.Freq;
Mag   = data.parameters.Mag;
Phase = data.parameters.Phase;

nFreq = length(Freq);
nMag  = length(Mag);
nPhase = length(Phase);

stabProp = data.stabilityproperties.stabProp;

switch action
    case '3D'
        
        stabChange = 0*stabProp;
        
        for iW = 1:1:nFreq
            for iM = 1:1:nMag
                for iP = 1:1:nPhase
                    if(stabProp(iW,iM,iP))
                        if( iM > 1 && iP > 1 && iM < nMag && iP < nPhase)
                            if(~stabProp(iW,iM-1,iP) || ~stabProp(iW,iM+1,iP))
                                stabChange(iW,iM,iP) = 1;
                            end
                            if(~stabProp(iW,iM,iP-1) || ~stabProp(iW,iM,iP+1))
                                stabChange(iW,iM,iP) = 1;
                            end
                        end
                    end
                end
            end
        end
        
        data.stabilityproperties.stabChange = stabChange;
        
        
    case '2D'
        
        % Initialization
        crit = zeros(nMag,nPhase);
        
        MPdata = struct;
        MPdata.NTR = {};
        MPdata.Mlow = {};
        MPdata.Plow = {};
        MPdata.Mupp = {};
        MPdata.Pupp = {};
        MPdata.Nst = {};
        MPdata.rev = {};
        
        for iW = 1:1:nFreq
            % Every iteration is one Magnitude-Phase per Frequency grid
            crit(:,:) = stabProp(iW,:,:);
            MPdata.NTR{iW} = [0 0];
            
            % Determine approach and initialization variables
            if max(max(crit(:,:))) == 0
                % there are no stable/passive combinations
                apprch = 2;
                MPdata.rev{iW} = -2;
                MPdata.NTR{iW} = [1 1];
            elseif min(min(crit(:,:))) == 1
                % there are only stable/passive combinations
                apprch = 2;
                MPdata.rev{iW} = -1;
                MPdata.NTR{iW} = [1 1];
                
            elseif max(min(crit)) == 0 && max(min(crit')) == 0;
                % No unconditionally stable/passive combinations:
                % if this is true, there are no columns/rows that exist entirely
                % of ones. min(crit) and min(crit') provide vectors with the
                % smallest value of each column and row respectively. So if there
                % are only ones, the smallest value is 1, and the maximum value of
                % these vectors will be one. So, if there are no rows or columns
                % existing entirely of ones, max(min(crit))=0 and
                % max(min(crit')) = 0. These columns/rows are necessary for
                % boundary estimation). Therefore, the approach is reversed, this
                % is shown in the third line!
                apprch = 1;
                MPdata.rev{iW} = 1;
                crit = crit*-1+1; % This is where the solution is reversed,
                % through this calculation, all zeros become
                % ones and all ones become zeros. Therefore,
                % instead of bounding the unstable/active area,
                % the algorithm is going to bound the
                % stable/passive area.
                NTR = [0 0];
            else
                apprch = 0;
                MPdata.rev{iW} = 0;
                NTR = [0 0];
            end
            
            if apprch < 2
                MPdata = initialMethod(iW,crit,nMag,nPhase,Mag,Phase,NTR,MPdata,MPdata.rev{iW});
            else
                MPdata.Mlow{iW} = [];
                MPdata.Plow{iW} = [];
                MPdata.Mupp{iW} = [];
                MPdata.Pupp{iW} = [];
            end
            
        end
        
        data.stabilityproperties.NTR = MPdata.NTR;
        data.stabilityproperties.Mlow = MPdata.Mlow;
        data.stabilityproperties.Plow = MPdata.Plow;
        data.stabilityproperties.Mupp = MPdata.Mupp;
        data.stabilityproperties.Pupp = MPdata.Pupp;
        data.stabilityproperties.Nst = MPdata.Nst;
        data.stabilityproperties.rev = MPdata.rev;
        
    case 'advanced'
        
        stableMag = {};
        unstableMag = {};
        consMag = {};
        stablePhase = {};
        unstablePhase = {};
        consPhase = {};
        noRegions = {};
        fourCorners = {};
        rev = {};
        
        step = data.stepOfAction(1,1);
        nostep = data.stepOfAction(1,2);
        
        for iW = 1:1:nFreq
            
            set(data.handles.defineElement,'String',...
                ['Mapping: 2D Bode diagram (f = ',num2str(Freq(iW)),' ',data.discretisation.unitsFreq.String{1},')  ',...
                '(Step ',num2str(step),' of ',num2str(nostep),').']); drawnow;
            
            MPpF(:,:) = stabProp(iW,:,:);
            stableMagPerFreq = [];
            unstableMagPerFreq = [];
            stablePhasePerFreq = [];
            unstablePhasePerFreq = [];
            consMagPerFreq = [];
            consPhasePerFreq = [];
            reversed = 0;
            
            if max(max(MPpF(:,:))) == 0
                % there are no stable/passive combinations
                noRegionsPerFreq = 1;
                unstableMagPerFreq = [Mag(1)-10,Mag(end)+10];
                unstablePhasePerFreq = [Phase(1)-10,Phase(end)+10];
                fourCornersOut = {[1 nPhase nMag 1]};
                fourCornersIn = {[1 nPhase nMag 1]};
                noRegionsPerFreq = 1;
                reversed = 0;
                
            elseif min(min(MPpF(:,:))) == 1
                % there are only stable/passive combinations
                noRegionsPerFreq = 0;
                stableMagPerFreq = [Mag(1)-10,Mag(end)+10];
                stablePhasePerFreq = [Phase(1)-10,Phase(end)+10];
                fourCornersOut = {[1 nPhase nMag 1]};
                fourCornersIn = {[1 nPhase nMag 1]};
                noRegionsPerFreq = 1;
                reversed = 0;
                
            elseif max(min(MPpF)) == 0 && max(min(MPpF')) == 0;
                % No unconditionally stable/passive combinations: approach should be
                % reversed
                MPpF = MPpF*-1+1;
                [MPpF,noRegionsPerFreq, fourCornersOut,fourCornersIn] = MPpFAnalysis(MPpF);
                
                [consMagPerFreq, stableMagPerFreq, unstableMagPerFreq, unstablePhasePerFreq, stablePhasePerFreq, consPhasePerFreq] =...
                    mergeRegions(Mag, Phase, fourCornersOut, fourCornersIn, 'reversed');
                reversed = 1;
                
            else
                [MPpF,noRegionsPerFreq, fourCornersOut,fourCornersIn] = MPpFAnalysis(MPpF);
                
                [consMagPerFreq, stableMagPerFreq, unstableMagPerFreq, unstablePhasePerFreq, stablePhasePerFreq, consPhasePerFreq] =...
                    mergeRegions(Mag, Phase, fourCornersOut, fourCornersIn, 'normal');
                reversed = 0;
                
            end
            
            noRegions{iW} = noRegionsPerFreq;
            fourCorners{iW} = {fourCornersOut,fourCornersIn};
            stableMag{iW} = stableMagPerFreq;
            unstableMag{iW} = unstableMagPerFreq;
            stablePhase{iW} = stablePhasePerFreq;
            unstablePhase{iW} = unstablePhasePerFreq;
            consMag{iW} = consMagPerFreq;
            consPhase{iW} = consPhasePerFreq;
            rev{iW} = reversed;
        end
        
        data.stabilityproperties.noRegions = noRegions;
        data.stabilityproperties.fourCorners = fourCorners;
        data.stabilityproperties.stableMag = stableMag;
        data.stabilityproperties.unstableMag = unstableMag;
        data.stabilityproperties.stablePhase = stablePhase;
        data.stabilityproperties.unstablePhase = unstablePhase;
        data.stabilityproperties.consMag = consMag;
        data.stabilityproperties.consPhase =  consPhase;
        data.stabilityproperties.rev = rev;
        
        step = step + 1;
        data.stepOfAction(1,1) = step;
        
        
end

return

function [consMagPerFreq, stableMagPerFreq, unstableMagPerFreq, unstablePhasePerFreq, stablePhasePerFreq, consPhasePerFreq] =...
    mergeRegions(Mag, Phase, fourCornersOut, fourCornersIn, action)

stableMagPerFreq = [];
unstableMagPerFreq = [];
stablePhasePerFreq = [];
unstablePhasePerFreq = [];
consMagPerFreq = [];
consPhasePerFreq = [];

noRegionsPerFreq = length(fourCornersOut);

for iNoReg = 1:1:noRegionsPerFreq
    inI = fourCornersIn{iNoReg};
    outI = fourCornersOut{iNoReg};
    
    if(noRegionsPerFreq > 1)
        consMagPerFreq = [Mag(1)-10, Mag(end)+10];
        consPhasePerFreq = [Phase(1)-10, Phase(end)+10];
    else
        consMagPerFreq = [Mag(outI(4)),Mag(inI(4)),Mag(inI(3)),Mag(outI(3))];
        consPhasePerFreq = [Phase(outI(1)), Phase(inI(1)), Phase(inI(2)), Phase(outI(2))];
    end
    
    switch action
        case 'normal'
            
            if(iNoReg == 1)
                unstableMagPerFreq = [Mag(inI(4)), Mag(inI(3))];
                unstablePhasePerFreq = [Phase(inI(1)), Phase(inI(2))];
                stableMagPerFreq = [Mag(1)-10,Mag(outI(4)),Mag(outI(3)),Mag(end)+10];
                stablePhasePerFreq = [Phase(1)-10,Phase(outI(1)),Phase(outI(2)),Phase(end)+10];
            else
                unstableMagPerFreq = checkUnstableRange(unstableMagPerFreq,Mag(inI(4)),Mag(inI(3)));
                unstablePhasePerFreq = checkUnstableRange(unstablePhasePerFreq,Phase(inI(1)),Phase(inI(2)));
                stableMagPerFreq =...
                    checkStableRange(stableMagPerFreq,[Mag(1)-10,Mag(outI(3))],[Mag(outI(4)),Mag(end)+10]);
                stablePhasePerFreq =...
                    checkStableRange(stablePhasePerFreq,[Phase(1)-10, Phase(outI(2))],[Phase(outI(1)),Phase(end)+10]);
            end
            
        case 'reversed'                  
            
            if(iNoReg == 1)
                stableMagPerFreq = [Mag(inI(4)), Mag(inI(3))];
                stablePhasePerFreq = [Phase(inI(1)), Phase(inI(2))];
                unstableMagPerFreq = [Mag(1)-10,Mag(outI(4)),Mag(outI(3)),Mag(end)+10];
                unstablePhasePerFreq = [Phase(1)-10,Phase(outI(1)),Phase(outI(2)),Phase(end)+10];
            else
                stableMagPerFreq = checkUnstableRange(stableMagPerFreq,Mag(inI(4)),Mag(inI(3)));
                stablePhasePerFreq = checkUnstableRange(stableMagPerFreq,Phase(inI(1)),Phase(inI(2)));
                unstableMagPerFreq =...
                    checkStableRange(unstableMagPerFreq,[Mag(1)-10,Mag(outI(3))],[Mag(outI(4)),Mag(end)+10]);
                unstablePhasePerFreq =...
                    checkStableRange(unstablePhasePerFreq,[Phase(1)-10, Phase(outI(2))],[Phase(outI(1)),Phase(end)+10]);
            end
    end
end

return

function sNew = checkStableRange(sOld,low,upp)

sNew = [];
for iNew = 1:1:2
    new1 = low(iNew);
    new2 = upp(iNew);
    for iOld = 1:2:length(sOld)-1
        old1 = sOld(iOld);
        old2 = sOld(iOld + 1);
        
        if(new1 <= old1)
            if(new2 <= old1)
                sNew = sNew;
            elseif(new2 > old1 && new2 <= old2)
                sNew = [sNew, old1, new2];
            elseif(new2 > old2)
                sNew = [sNew, old1, old2];
            end
        elseif(new1 > old1  && new1 <= old2)
            if(new2 > old1 && new2 <= old2)
                sNew = [sNew, new1, new2];
            elseif(new1 > 1 && new2 > old2 && new2 <= old2)
                sNew = [sNew, new1, old2];
            elseif(new1 > old2)
                sNew = sNew;
            end
        elseif(new1 <= old1 && new2 > old2)
            sNew = [sNew, old1, old2];
        end
    end
end

return

function sNew = checkUnstableRange(sOld,new1,new2)

sNew = [];

for iOld = 1:2:length(sOld)-1   
    old1 = sOld(iOld);
    old2 = sOld(iOld + 1);
    if(new1 <= old1)
        if(new2 <= old1)
            %1
            sNew = [new1,new2,old1,old2];
        elseif(new2 > old1 && new2 <= old2)
            %2
            sNew = [old1,new2];
        else
            %6
            sNew = [old1,old2];
        end
    elseif(new1 > old1 && new1 <= old2)
        if(new2 <= old2)
            %3
            sNew = [new1, new2];
        else
            %4
            sNew = [new1, old2];
        end
    else
        %5
        sNew = [old1, old2, new1, new2];
    end
    
end

return

function MPdata = initialMethod(iW,crit,nMag,nPhase,Mag,Phase,NTR,MPdata,rev)

Plow = []; Pupp = []; Mlow = []; Mupp = [];
Nst = 0;

%% Find moduli range in which all phases are passive
% Calculate boundaries (Note: crit(M,P))
% Stops when leftbnd end rightbnd is found
TrM = min(crit');
% This is a vector that evaluates the smallest value
% for each row (magnitude). If this is 1, this
% particular magnitude is stable/passive for any
% phase!
ddM = 0;        % Size to compensate if the bound is on boundary
iMlow = 1;      % Index for lower bound
iMupp = nMag;   % Index for upper bound

% This part below lowers the lower bound or increased the upper bound
% if the bound is at the boundary of the crit-matrix.

% leftbnd and rightbnd are indices whether the bound is found (0) or
% not (1).
if TrM(1)
    % first magnitude (row) is stable/passive for any phase
    leftbnd = 1;
    Mlow(1,1) = Mag(iMlow);
else
    % unstable/active area is already at first magnitude, i.e. the left
    % bound is found!
    leftbnd = 0;
    Mlow(1,1) = Mag(iMlow)/2; % Then the bound for this area will be
    % lower then Mag(iMllow) (division by 2, could also be division by 3,4,5,
    % doesn't matter, as long as the result is lower than Mag(iMlow)). For plotting
    % reasons only (!!) this value is lowered (otherwise the bound would be
    % visible at the first value while in fact this is not the actual bound).
    ddM = ddM+1;
end

% This is the same strategy as above.
if TrM(end)
    rightbnd = 1;
    Mupp(1,1) = Mag(iMupp);
else
    rightbnd = 0;
    Mupp(1,1) = 2*Mag(iMupp);
    ddM = ddM+1;
end

% Check inner stability grid
for iM = 2:nMag
    if TrM(iM) && leftbnd
        % magnitude Mag(iM) is stable/passive for all phase points and
        % left bound is not found yet
        iMlow = iM;% lower bound is increased
        Mlow(1,1) = Mag(iMlow);
    else
        % magnitude Mag(iM) is not stable/passive for all phase points,
        % so this is the lower magnitude boundary!
        leftbnd = 0;
    end
    % Similar strategy for right bound
    if TrM(end+1-iM) && rightbnd
        iMupp = nMag+1-iM;
        Mupp(1,1) = Mag(iMupp);
    else % bound is found
        rightbnd = 0;
    end
end
dM = iMupp-iMlow-1+ddM; % This is the size of the unstable region
% in terms of magnitude (the ddM is to
% compensate for the fact that the boundary
% is a bit larger because the boundary is at
% the minimum or maximum magnitude that in Mag occurs.

%% Check phase range (possibly more phases like explained up top in dutch)
% Calculate boundaries (Note: crit(M,P))
TrP = min(crit);% Find for every column (phase) the smallest value

if TrP(1)
    % The first phase is stable for all magnitude points
    Ntr = 0;% No unstable region found yet (Ntr = number of unstable phase regions!)
else
    Ntr = 1;% Unstable region found!
    iPlow = 1;% Lower boundary of phase
    Plow(1,1) = Phase(1)-10;% Define lower bound outside grid
    % Similar to magnitude -> lowering the phase
    % because the actual boundary is lower than
    % Phase(1).
end

for iP = 2:nPhase
    if ~TrP(iP-1) && TrP(iP)
        % If previous phase point was unstable and current phase point is
        % stable (for all magnitudes) then the upper bound is found!
        iPupp = iP;
        Pupp(Ntr,1) = Phase(iPupp);
    elseif TrP(iP-1) && ~TrP(iP)
        % If previous phase point was stable and current phase point is
        % unstable (for all magnitudes) then the lower bound is found!
        Ntr = Ntr+1;% New unstable phase region is found!
        iPlow = iP-1;
        Plow(Ntr,1) = Phase(iPlow);
    end
end

if ~TrP(end)
    iPupp(Ntr) = nPhase;
    Pupp(Ntr,1) = Phase(end)+10;
end

NTR(1) = Ntr;% Number of phase regions
dP = sum(~TrP);
ARout = dM*dP;% Total number of points inside the unstable region.

%% For each connected region, find largest active magnitude/ phase range
%% inside passive magnitude/ phase range
% Up to now, the bounds on the unstable region is found! WIth this
% information we can plot the dark grey areas! Second step begins now:
% to find a rectangle that only bounds unstable points: with this info
% we can plot the white areas.
ARinn = 0;
% This is how we are going to find this white area: we want to bound
% as much points as possible. This amount is ARinn. The boundaries
% that give us the highest ARinn is going to be chosen! This is the
% least conservative solution.
crit = crit*-1+1;
for iMl = iMlow:iMupp
    % We are going to move through all possible magnitude points and
    % evaluate which combination of magnitude boundaries include the most
    % points!
    for iMu = iMlow:iMupp
        if iMl <= iMu
            W = iMu-iMl+1;
            if W > 1
                % Lower boundary should be lower than upper boundary
                H = sum(min( crit(iMl:iMu,:) ));
                % This is the height of the unstable region, i.e. the
                % number of phase points that are stable between iMl
                % and iMu.
            else
                H = sum( crit(iMl:iMu,:) );
            end
            ARnew = W*H;% Number of points inside rectangle is width times height
        else
            ARnew = 0;
        end
        if ARnew > ARinn;
            % Change parameters if newly acquired rectangle is better than
            % previous one (more points inside newly acquired rectangle)
            iMlow2 = iMl;
            iMupp2 = iMu;
            if iMl == 1
                Mlow(1,2) = Mag(1)/2;
            else
                Mlow(1,2) = Mag(iMl);
            end
            if iMu == nMag
                Mupp(1,2) = Mag(end)*2;
            else
                Mupp(1,2) = Mag(iMu);
            end
            ARinn = ARnew;
        end
    end
end

% Nst is a measure for the amount of conservatism added. If Nst = 100%,
% then ARinn = ARout: there is no uncertainty and the outer bounds
% equal the inner bounds!
Nst = ARinn/ARout*100;
%% Check phase range (possibly more phases)
% The phase has already been used to determine the rectangle. This
% step simply evaluates how many phase ranges exist and allocates
% them appropriately.
if iMlow2 < iMupp2
    TrP = min(crit(iMlow2:iMupp2,:));
else
    TrP = crit(iMlow2:iMupp2,:);
end

if TrP(1)
    Ntr = 1;
    iPlow = 1;
    Plow(1,2) = Phase(1)-10;
else
    Ntr = 0;
end

for iP = 2:nPhase
    if TrP(iP-1) && ~TrP(iP)
        iPupp = iP-1;
        Pupp(Ntr,2) = Phase(iPupp);
    elseif ~TrP(iP-1) && TrP(iP)
        Ntr = Ntr+1;
        iPlow = iP;
        Plow(Ntr,2) = Phase(iPlow);
    end
end
if TrP(end)
    iPupp(Ntr) = nPhase;
    Pupp(Ntr,2) = Phase(end)+10;
end
NTR(2) = Ntr;% Number of phase regions in inner rectangle!


MPdata.NTR{iW} = NTR;
MPdata.Mlow{iW} = Mlow;
MPdata.Plow{iW} = Plow;
MPdata.Mupp{iW} = Mupp;
MPdata.Pupp{iW} = Pupp;
MPdata.Nst{iW} = Nst;
MPdata.rev{iW} = rev;
return

function [B,noRegions,fourCornersOut,fourCornersIn] = MPpFAnalysis(A)

label = 0;
[n,m] = size(A);

% Z = [ones(1,m+2);ones(n,1),A,ones(n,1);ones(1,m+2)];
Z = A;
B = zeros(n,m);
for i = 1:1:n
    for j = 1:1:m
        if(Z(i,j) == 1)
            B(i,j) = 0;
        elseif(Z(i,j)==0)
            if(j == 1)
                if(i == 1)
                    label = label + 1;
                    B(i,j) = label;
                else
                    if(B(i-1,j) ~= 0)
                        B(i,j) = B(i-1,j);
                    else
                        label = label + 1;
                        B(i,j) = label;
                    end
                end
            else
                if(i == 1)
                    if(B(i,j-1) ~= 0)
                        B(i,j) = B(i,j-1);
                    else
                        label = label + 1;
                        B(i,j) = label;
                    end
                else
                    % it's not the first row and not the first column
                    upper = B(i-1,j);%upper
                    leftupper = B(i-1,j-1);%leftupper
                    left = B(i,j-1);%left
                    if(~upper)
                        if(~leftupper)
                            if(~left)
                                label = label + 1;
                                B(i,j) = label;
                            else
                                B(i,j) = left;
                            end
                        else
                            B(i,j) = leftupper;
                            if(left)
                                B = substituteLabels(B,left,leftupper,i,j);
                            end
                        end
                    else
                        if(~leftupper)
                            B(i,j) = upper;
                            if(left)
                                B = substituteLabels(B,left,upper,i,j);
                            end
                        else
                            B(i,j) = leftupper;
                            if(left)
                                B = substituteLabels(B,upper,leftupper,i,j);
                            end
                        end
                    end
                end
            end
        end
    end
end

maxElement = max(max(B));
noRegions = 0;
if(maxElement > 0)
    for iB = 1:1:maxElement
        if(~isempty(find(B == iB)))
            noRegions = noRegions + 1;
            label(noRegions) = iB;
        end
    end
end

unstableBoundary = bwboundaries(B,8);

noAreas = length(unstableBoundary);
areaIn = [];
fourCorners = [];
lowXout = [];
lowYout = [];
uppXout = [];
uppYout = [];

for i = 1:1:noAreas
    bound = unstableBoundary{i};
    
    % Outer
    lowXout = min(bound(:,2)-1);
    lowYout = max(bound(:,1)+1);
    uppXout = max(bound(:,2)+1);
    uppYout = min(bound(:,1)-1);
    
    fourCornersOut{i} = [lowXout,uppXout,lowYout,uppYout];
    
    % Inner
    areaIn(i) = 0;
    fourCornersIn{i} = 0;
    
    for j = 1:1:length(bound)
        
        y = bound(j,1);
        x = bound(j,2);
        % Check possibilities
        % Go right
        xLow = x;
        xUp = x;
        yUp = y;
        yLow = y+1;
        
        [areaIn(i),fourCornersIn{i}] = areaComputation(xLow,xUp,yLow-1,yUp,areaIn(i),fourCornersIn{i});
        
        xi = xLow;
        yi = yLow-1;
        
        [~,q] = size(Z);
        while(q > xi && Z(yi,xi) == 0)
            if(Z(yi,xi) == 0)
                xUp = xi;
                [areaIn(i),fourCornersIn{i}] = areaComputation(xLow,xUp,yLow-1,yUp,areaIn(i),fourCornersIn{i});
                xi = xi + 1;
            end
        end
        yi = yi + 1;
        xi = xUp;
        
        [q,~] = size(Z);
        while(xUp >= xLow && yLow > yUp && yLow < q)
            if(max(max(Z(yUp:yLow+1,xLow:xUp))) == 0)
                yLow = yLow + 1;
                [areaIn(i),fourCornersIn{i}] = areaComputation(xLow,xUp,yLow,yUp,areaIn(i),fourCornersIn{i});
            else
                xUp = xUp - 1;
            end
        end
        
    end
    
    for k = 1:1:4
        if(k == 1 || k == 4)
            if(fourCornersIn{i}(k) < 1)
                fourCornersIn{i}(k) = 1;
            end
            if(fourCornersOut{i}(k) < 1)
                fourCornersOut{i}(k) = 1;
            end
        elseif(k == 2)
            if(fourCornersIn{i}(k) > m)
                fourCornersIn{i}(k) = m;
            end
            if(fourCornersOut{i}(k) > m)
                fourCornersOut{i}(k) = m;
            end
        else
            if(fourCornersOut{i}(k) > n)
                fourCornersOut{i}(k) = n;
            end
            if(fourCornersIn{i}(k) > n)
                fourCornersIn{i}(k) = n;
            end
        end
    end
    
end

return

function B = substituteLabels(B,label,element,i,j)

for k = 1:1:i
    for l = 1:1:j
        if(B(k,l) == label)
            B(k,l) = element;
        end
    end
end

return

function [areaIn, fourClockwiseCorners] = areaComputation(xLow,xUp,yLow,yUp,areaIn,fourClockwiseCorners)

areaInNew = (xUp-xLow+1)*(yLow-yUp+1);
if(areaInNew > areaIn)
    areaIn = areaInNew;
    fourClockwiseCorners = [xLow xUp yLow yUp];
end

return
